import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.util.*;

/**
 * A user-controller checkers piece.
 * Use the mouse to drag the piece to its destination.
 * When the mouse is released the move is checked and
 * implemented. Attempts to move the piece to an illegal
 * location will cause the position to be reset.
 * 
 * @author David J. Barnes
 * @version 2008.08.05
 */
public class DarkMan extends Man
{
    // Is the mouse currently being dragged on this piece?
    private boolean beingDragged;
    // If the mouse is being dragged, where it started.
    private int dragStartX, dragStartY;
    // If the mouse is being dragged, a list of the locations
    // it has been dragged through.
    private LinkedList<Location> dragPath;

    /**
     * Indicate the color of piece to use, and the opponent type.
     */
    public DarkMan(Class opponent)
    {
        super(Color.RED, opponent);
        beingDragged = false;
        dragPath = new LinkedList<Location>();
    }
    
    /**
     * The move is supplied by the user using the mouse.
     * Drag through the squares to be moved through.
     */
    public void act() 
    {
        super.act();
    }
    
    /**
     * Choose a path representing a move from the start location to
     * the end location, including any intermediate squares.
     * @return The chosen path, or null if none is chosen.
     */
    public List<Location> chooseMovePath()
    {
        // The path to be returned, if one is chosen.
        // One will only be chosen once the mouse is released.
        List<Location> path = null;
        
        if(Greenfoot.mouseDragged(this)) {            
            int x = getX(), y = getY();
            if(!beingDragged) {
                recordStartOfMove();
            }
            // Record the location moved to.
            moveToMouse(Greenfoot.getMouseInfo());            
        }
        
        if(Greenfoot.mouseDragEnded(this)) {
            // Implement the move to the current square and
            // then check the complete move.
            if(moveToMouse(Greenfoot.getMouseInfo())) {
                if(isLegal(dragPath)) {
                    path = new LinkedList<Location>(dragPath);
                }
                else {
                    resetLocation();
                }
            }
            else {
                resetLocation();
            }
            beingDragged = false;
            dragPath.clear();
        }
        
        return path;
    }
    
    /**
     * Check for a diagonal move in the right direction.
     * Kings may move both forwards and backwards.
     */
    protected boolean dragOk(MouseInfo mouse)
    {
        if(Math.abs(mouse.getX() - getX()) == Math.abs(mouse.getY() - getY())) {
            // The basic movement is ok.
            if(isKing()) {
                // Kings can move in any direction.
                return true;
            }
            else {
                // This piece must move up the board.
                return mouse.getY() < getY();
            }
        }
        else {
            return false;
        }
    }

    /**
     * Record the start of a move.
     */
    private void recordStartOfMove()
    {
        int x = getX(), y = getY();
        beingDragged = true;
        dragStartX = x;
        dragStartY = y;
        dragPath.clear();
        dragPath.add(new Location(x, y));
    }
    
    /**
     * Try to enact a move to the given mouse location.
     * @return True if the move is ok, false otherwise.
     */
    private boolean moveToMouse(MouseInfo mouse)
    {
        int newX = mouse.getX(), newY = mouse.getY();
        if(dragOk(mouse)) {
            moveTo(newX, newY);
            return true;
        }
        else if(newX == dragStartX && newY == dragStartY) {
            // Allow return to the original location.
            return false;
        }
        else if(newX == getX() && newY == getY()) {
            return true;
        }
        else {
            // Illegal location.
            return false;
        }
    }
    
    /**
     * Record a move to the given location.
     * @param x The x-coordinate of the location.
     * @param y The y-coordinate of the location.
     */
    private void moveTo(int x, int y)
    {
        if(dragPath.size() > 0 ) {
            Location lastLocation = dragPath.getLast();
            if(lastLocation.getX() != x || lastLocation.getY() != y) {
                dragPath.add(new Location(x, y));
            }
        }
        else {
            dragPath.add(new Location(x, y));
        }
        setLocation(x, y);
    }
    
    /**
     * Reset back to the starting location and clear the path.
     */
    private void resetLocation()
    {
        setLocation(dragStartX, dragStartY);
        dragPath.clear();
    }
    
}
