import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Font;
import java.util.*;

/**
 * The common elements of a checker piece.
 * 
 * The act method of subclasses should arrange to identify the chosen move via
 * a call to chooseMovePath and setMoveAvailable. The chooseMovePath must be
 * implemented in the subclass to provide a particular selection technique - e.g.
 * mouse selection or material selection.
 * 
 * The move path consists of a list of Locations passed through by the moving piece,
 * including the start and ending squares and any capture squares along the way.
 * 
 * A selected move for just one actor is enacted via a call to the play method, invoked by the
 * Board's act method.
 * 
 * @author David J. Barnes
 * @version 2008.08.06
 */
public abstract class Man extends Actor
{
    // The class of this piece's opponent.
    private Class opponent;
    // Is this piece a king?
    private boolean isKing;
    // Whether this piece has a move available or not.
    private boolean moveAvailable;
    // The available move, if moveAvailable.
    private List<Location> movePath;
    
    /**
     * Set up this piece with the given color and opponent.
     * @param color The image color for this piece.
     * @param opponent The opponent for this piece.
     */
    public Man(Color color, Class opponent)
    {
        this.opponent = opponent;
        isKing = false;
        
        moveAvailable = false;
        movePath = new LinkedList<Location>();
        
        int cellsize = Board.CELLSIZE;
        int mansize = cellsize - 8;
        GreenfootImage image = new GreenfootImage(mansize, mansize);
        image.setColor(color);
        image.fillOval(0, 0, mansize, mansize);
        setImage(image);
    }
    
    /**
     * The default implementation of a piece. If it is this piece's turn
     * then choose a path for it, and make it available. There may be no
     * path available for the piece.
     */
    public void act() 
    {
        // Make sure it is this piece's turn.
        if(myTurn()) {
            // Try to choose a move - there may be none.
            List<Location> path = chooseMovePath();
            if(path != null) {
                // Record that a move is available.
                setMoveAvailable(path);    
            }
        }
        else {
            resetMoveAvailable();
        }        
    }

    /**
     * Choose a path representing a move from the piece's location to
     * its final location, including any intermediate squares.
     * @return The chosen path, or null if none is chosen or available.
     */
    abstract public List<Location> chooseMovePath();
    
    /**
     * Return whether this piece is a king or not.
     * @return true if it is a king, false otherwise.
     */
    public boolean isKing()
    {
        return isKing;
    }
    
    /**
     * Whether this piece has a move available to be played.
     * @return true If there is a move available, false otherwise.
     */
    public boolean isMoveAvailable()
    {
        return moveAvailable;
    }
    
    /**
     * Clear the move availability.
     */
    public void resetMoveAvailable()
    {
        moveAvailable = false;
        movePath.clear();
    }
    
    /**
     * Return the move path, if any.
     * @return The move path.
     */
    public List<Location> getMovePath()
    {
        return movePath;
    }    
    
    /**
     * Play the available move. If it is played successfully then
     * the sides are switched ready for the opponent's pieces.
     */
    public boolean play()
    {
        if(isMoveAvailable() && makeCompleteMove(movePath)) {
            resetMoveAvailable();
            ((Board) getWorld()).switchSides();
            return true;
        }
        else {
            return false;
        }
    }
    
    /**
     * Determine whether it is this type of piece's turn to move.
     * @return true If it is this piece's turn to move, false otherwise.
     */
    public boolean myTurn()
    {
        Board board = (Board) getWorld();
        return getClass().equals(board.getCurrentTurn());
    }
    
    /**
     * Enact the move through the locations listed in the path.
     * @param path The locations moved through, including the starting and
     *             ending location.
     * @return true If the move was ok, false otherwise.
     */
    public boolean makeCompleteMove(List<Location> path)
    {
        Board board = (Board) getWorld();
        if(!getClass().equals(board.getCurrentTurn())) {
            return false;
        }
        if(path.size() < 2) {
            return false;
        }
        else{
            // Check that the move was legal.
            if(isLegal(path)) {
                makeCaptures(path);
                Location finalLocation = path.get(path.size() - 1);
                setLocation(finalLocation.getX(), finalLocation.getY());
                if(!isKing()) {
                    checkForBecomingKing();
                }
                return true;
            }
            else {
                return false;
            }
        }
    }
    
    /**
     * Check whether the path is legal - either a single move
     * or one or more captures.
     * @param path The path to be checked.
     * @return true if the path is legal, false otherwise.
     */
    public boolean isLegal(List<Location> path)
    {
        World world = getWorld();
        if(path.size() < 2) {
            return false;
        }
        else if(!pathValid(path)) {
            return false;
        }
        else if(path.size() == 2) {
            // Should be a move to an empty location.
            Location finalLocation = path.get(path.size() - 1);
            List<Actor> actors = world.getObjectsAt(finalLocation.getX(), finalLocation.getY(),
                                                    Actor.class);
            actors.remove(this);
            return actors.size() == 0;
        }
        else {
            // Hop over captured pieces to a final empty location.
            Iterator<Location> it = path.iterator();
            // Drop the first, which is the starting square.
            it.next();
            // Whether we are looking for a capture or a free square.
            boolean findCapture = true;
            while(it.hasNext()) {
                Location loc = it.next();
                List<Actor> captured =
                   world.getObjectsAt(loc.getX(), loc.getY(), opponent);
                if(findCapture) {
                    // There should be a piece here.
                    if(captured.size() == 1) {
                        // Ok. Look for a free square next.
                        findCapture = false;
                    }
                    else {
                        return false;
                    }
                }
                else {
                    // Should be an empty square.
                    if(captured.size() == 0) {
                        // Ok. Look for a capture next.
                        findCapture = true;
                    }
                    else {
                        return false;
                    }
                }
            }
            return true;
        }
    }
    
    /**
     * Return a string representation of this piece.
     * @return A string representation of this piece.
     */
    public String toString()
    {
        return "(" + getX() + "," + getY() + ") " +
               moveAvailable + " " + movePath;
    }
    
    /**
     * Set the given move path as a possible move for this piece.
     * @param path The move's path.
     */
    protected void setMoveAvailable(List<Location> path)
    {
        movePath.clear();
        movePath.addAll(path);
        moveAvailable = true;
    }
    
    /**
     * Return the opponent's class.
     * @return The opponent's class.
     */
    protected Class getOpponent()
    {
        return opponent;
    }
    
    /**
     * Sanity check on the moves in the path.
     */
    protected boolean pathValid(List<Location> path)
    {
        // Check that all locations are within the board.
        for(Location loc : path) {
            int x = loc.getX(), y = loc.getY();
            if(x < Board.MIN_X || x > Board.MAX_X) {
                return false;
            }
            if(y < Board.MIN_Y || y > Board.MAX_Y) {
                return false;
            }
        }
        // Check that all moves are adjacent diagonal moves.
        Iterator<Location> it = path.iterator();
        Location current = it.next();
        while(it.hasNext()) {
            Location next = it.next();
            if(Math.abs(current.getX() - next.getY()) != 1 &&
               Math.abs(current.getY() - next.getY()) != 1) {
                   return false;
            }
            current = next;
        }
        return true;
    }
    
    /**
     * Make the captures along the path.
     * This method assums that the path has already been
     * checked for legality.
     */
    private void makeCaptures(List path)
    {
        assert isLegal(path);
        if(path.size() > 2) {
            World world = getWorld();
            Iterator<Location> it = path.iterator();
            // Drop the first, which is the starting square.
            it.next();
            // Whether we are looking for a capture or a free square.
            boolean findCapture = true;
            // The next square of interest.
            Location loc = it.next();
            // The last location is the destination square so we
            // don't need to check it.
            while(it.hasNext()) {
                // Get any piece that is on this square.
                List<Actor> captured =
                   world.getObjectsAt(loc.getX(), loc.getY(), opponent);
                if(findCapture) {
                    // Ok. Look for a free square next.
                    world.removeObject(captured.get(0));
                    findCapture = false;
                }
                else {
                    // Ok. Look for a capture next.
                    findCapture = true;
                }
                loc = it.next();
            }
        }
    }
    
    /**
     * Make this piece a king.
     */
    private void makeKing()
    {
        isKing = true;
        makeKingImage();
    }
    
    /**
     * Check whether this piece has reach the end of the board
     * and become a king.
     */
    private void checkForBecomingKing()
    {
        if(getY() == 0 || getY() == 7) {
            makeKing();
        }
    }
    
    /**
     * Create the image to represent a piece that is a king.
     */
    private void makeKingImage()
    {
        GreenfootImage image = new GreenfootImage(getImage());
        image.setColor(Color.BLACK);
        Font font = image.getFont();
        image.setFont(font.deriveFont(24.0F));  // use larger font
        image.drawString("K", image.getWidth() / 2 - 8, image.getHeight()/ 2 + 10);
        setImage(image);
    }
            
}
