import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.util.*;

/**
 * A piece that selects its longest capture sequence in preference
 * to a simple move to a free square.
 * There is no intelligence in the move selection.
 * 
 * @author David J. Barnes
 * @version 2008.08.06
 */
public class ThinkingMan extends Man
{
    // This piece naturally moves down the board, until it becomes a king.
    private static final int NATURAL_DIRECTION = 1;
    
    /**
     * Create a thinking man with the given opponent.
     * @param opponent This man's opponent.
     */
    public ThinkingMan(Class opponent)
    {
        super(Color.WHITE, opponent);
    }
    
    /**
     * Identify a move, if possible.
     * Prefer a long capture over a single move.
     */
    public void act() 
    {
        super.act();
    }
    
    /**
     * Choose a path representing a move from the start location to
     * the end location, including any intermediate squares.
     * @return The chosen path, or null if none is chosen.
     */
    public List<Location> chooseMovePath()
    {
        int x = getX(), y = getY();
        // Find a possible path. There may be none.
        List<Location> path = null;
        
        // Get a list of all adjacent locations.
        List<Location> adjacentLocations = getAdjacentLocations(x, y);
        // Look for adjacent locations with an opponent's piece.
        List<Location> possibleCaptureLocations = getCaptureLocations(adjacentLocations);
        if(possibleCaptureLocations.size() != 0) {
            // See if any of the opponent's pieces can be captured.
            path = getCapturePath(possibleCaptureLocations, x, y);
        }
        if(path == null) {
            // Still need a move.
            // Look for moves into an adjacent free location.
            List<Location> locations = getSingleMoves(adjacentLocations);
            if(locations.size() > 0) {
                // Pick one at random.
                path = new LinkedList<Location>();
                path.add(new Location(x, y));
                path.add(locations.get(Greenfoot.getRandomNumber(locations.size())));
            }
        }
        return path;
    }
    
    /**
     * Return a list of locations adjacent to the current
     * location, that contain an opponent's piece.
     */
    private List<Location> getCaptureLocations(List<Location> locations)
    {
        List<Location> captureLocations = new LinkedList<Location>();
        Iterator<Location> it = locations.iterator();
        while(it.hasNext()) {
            Location loc = it.next();
            if(getWorld().getObjectsAt(loc.getX(), loc.getY(), 
                    getOpponent()).size() > 0) {
                captureLocations.add(loc);
            }
        }
        return captureLocations;
    }
    
    /**
     * Recursively build a move from (x,y) that is a capture and return its path.
     * If there is no capture available, return null.
     * WARNING: If there is a circular path for a king around a group
     * of opponents pieces then this implementation will not terminate!
     * @param possibleCaptureLocations Possible capture locations.
     * @param x The x-coordinate of the launching point.
     * @param y The y-coordinate of the launching point.
     * @return The chosen path, or null if none found.
     */
    private List<Location> getCapturePath(List<Location> possibleCaptureLocations, 
                                          int x, int y)
    {
        // Isolate those moves that are actual captures.
        List<Location> captureLocations = new LinkedList<Location>();
        // Need to ensure that the location on the other side is free.
        Iterator<Location> it = possibleCaptureLocations.iterator();
        while(it.hasNext()) {
            Location loc = it.next();
            int freeX = loc.getX() + loc.getX() - x;
            int freeY = loc.getY() + loc.getY() - y;
            if(validX(freeX) && validY(freeY)) {
                if(freeLocation(freeX, freeY)) {
                    captureLocations.add(loc);
                }
            }
        }
        
        if(captureLocations.size() > 0) {
            // At least one capture is available. Build a path of the longest
            // capture sequence.
            List<Location> path = new LinkedList<Location>();
            for(Location loc : captureLocations) {
                // The free square beyond the capture square.
                int freeX = x + 2 * (loc.getX() - x);
                int freeY = y + 2 * (loc.getY() - y);
                // Potential further captures.
                List<Location> furtherLocations = getCaptureLocations(
                                    getAdjacentLocations(freeX, freeY));
                // Avoid spurious oscillations by a king back and forth over the
                // same piece.
                furtherLocations.remove(loc);
                if(furtherLocations.size() != 0) {
                    // Recursively explore this capture path.
                    List<Location> furtherPath = getCapturePath(
                            furtherLocations, freeX, freeY);
                    if(furtherPath != null) {
                        // See if this is longer than the existing one.
                        if(furtherPath.size() > path.size() - 1) {
                            // A longer path.
                            path.clear();
                            
                            path.add(loc);
                            path.addAll(furtherPath);
                        }
                    }
                }                    
            }
            if(path.size() == 0) {
                // No further captures beyond any of the current ones.
                // Just pick one at random.
                Location loc = captureLocations.get(
                                    Greenfoot.getRandomNumber(captureLocations.size()));
                path.add(loc);
                // The free square beyond the capture square.
                int freeX = x + 2 * (loc.getX() - x);
                int freeY = y + 2 * (loc.getY() - y);
                path.add(new Location(freeX, freeY));
            }
            // Add in the start of this part of the path.
            path.add(0, new Location(x, y));
            return path;
        }
        else {
            return null;
        }
    }
    
    /**
     * Return a list of empty locations adjacent to the current
     * location, that represent possible moves.
     */
    private List<Location> getSingleMoves(List<Location> locations)
    {
        List<Location> freeLocations = new LinkedList<Location>();
        for(Location loc : locations) {
            if(freeLocation(loc.getX(), loc.getY())) {
                freeLocations.add(loc);
            }
        }
        return freeLocations;
    }
    
    /**
     * Return a list of the adjacent locations. This takes into
     * account whether this piece is a king or not. It does not
     * matter whether the locations are empty or not.
     */
    private List<Location> getAdjacentLocations(int x, int y)
    {
        List<Location> moves = new LinkedList<Location>();
        int newY;
        
        // Can we move in the natural direction?
        newY = y + NATURAL_DIRECTION;
        if(validY(newY)) {
            adjacentXMoves(moves, x, newY);
        }
        if(isKing()) {
            newY = y - NATURAL_DIRECTION;
            if(validY(newY)) {
                adjacentXMoves(moves, x, newY);
            }
        }
        return moves;
    }
    
    /**
     * Store in moves the locations that are adjacent to the
     * current location in row y.
     * @param moves Where to store the adjacent moves.
     * @param x x-coordinate to be offset from.
     * @param y y-coordinate of the adjacent moves.
     */
    private void adjacentXMoves(List<Location> moves, int x, int y)
    {
        // Try left.
        int newX = x - 1;
        if(validX(newX)) {
            moves.add(new Location(newX, y));
        }
        // Try right.
        newX = x + 1;
        if(validX(newX)) {
            moves.add(new Location(newX, y));
        }
    }
    
    /**
     * @return true if the given location is free, false otherwise.
     */
    private boolean freeLocation(int x, int y)
    {
        return getWorld().getObjectsAt(x, y, Actor.class).size() == 0;
    }
    
    /**
     * Determine whether x is valid within the board.
     * @param x Possible x-coordinate.
     * @return true if x is valid, false otherwise.
     */
    private boolean validX(int x)
    {
        return x >= Board.MIN_X && x <= Board.MAX_X;
    }
    
    /**
     * Determine whether y is valid within the board.
     * @param y Possible y-coordinate.
     * @return true if y is valid, false otherwise.
     */
    private boolean validY(int y)
    {
        return y >= Board.MIN_Y && y <= Board.MAX_Y;
    }
}
